#pragma once

// [mem-buffer.h]
// -------------------------------------------------------------------------------------------- //
// fast memory buffer for c primary data types
// -------------------------------------------------------------------------------------------- //

#include "scm-mem.h"

#include <string.h>

namespace scm
{
	template< typename T >
	class mem_buffer {
	public:

		// append value
		// -------------------------------------------------------------------------------------------- //
		void append(const T value);
		void append(const T* value, const int n);
		// -------------------------------------------------------------------------------------------- //
		
		// reset
		// -------------------------------------------------------------------------------------------- //
		void reset();
		// -------------------------------------------------------------------------------------------- //

		// get calls
		// -------------------------------------------------------------------------------------------- //
		const T* get_ptr() const;
		T get_value(const int idx) const;
		int get_size() const;
		// -------------------------------------------------------------------------------------------- //

		// -------------------------------------------------------------------------------------------- //
		mem_buffer();
		~mem_buffer();
		// -------------------------------------------------------------------------------------------- //

	private:
		// data (private):
		// -------------------------------------------------------------------------------------------- //

		T *buf;
		int size;
		int nalloc;

		static const int c_nalloc_inc = 256;
	};
}
// -------------------------------------------------------------------------------------------- //

// -------------------------------------------------------------------------------------------- //
// * Implementation *
// -------------------------------------------------------------------------------------------- //

template< typename T >
scm::mem_buffer< T >::mem_buffer()
{
	size = 0;
	nalloc = 0;
}

template< typename T >
scm::mem_buffer< T >::~mem_buffer()
{
	if (nalloc > 0)
		deallocate(buf);

	nalloc = 0;
	size = 0;
}
// -------------------------------------------------------------------------------------------- //

// * append * //
// -------------------------------------------------------------------------------------------- //
template< typename T >
void scm::mem_buffer< T >::append(const T value)
{
	if (size == nalloc) {
		T *lbuf;
		allocate(&lbuf, nalloc + c_nalloc_inc);
		if (size > 0) {
			memcpy(lbuf, buf, size * sizeof(T));
			deallocate(buf);
		}

		buf = lbuf;
		nalloc += c_nalloc_inc;
	}

	buf[size] = value;
	size++;
}

template< typename T >
void scm::mem_buffer< T >::append(const T* value, const int n)
{
	if ((size - 1) + n >= nalloc) {
		T *lbuf;
		allocate(&lbuf, nalloc + max(c_nalloc_inc, size + n - nalloc));
		if (size > 0) {
			memcpy(lbuf, buf, size * sizeof(T));
			deallocate(buf);
		}

		buf = lbuf;
		nalloc += max(c_nalloc_inc, size + n - nalloc);
	}

	memcpy(&buf[size], value, n);
	size += n;
}
// -------------------------------------------------------------------------------------------- //

// * reset * //
// -------------------------------------------------------------------------------------------- //
template< typename T >
void scm::mem_buffer< T >::reset()
{
	size = 0;
}
// -------------------------------------------------------------------------------------------- //

// * get calls * //
// -------------------------------------------------------------------------------------------- //
template< typename T >
const T* scm::mem_buffer< T >::get_ptr() const
{
	return buf;
}

template< typename T >
T scm::mem_buffer< T >::get_value(const int idx) const
{
	if ((idx >= 0) && (idx < size))
		return buf[idx];

	return (T)0;
}

template< typename T >
int scm::mem_buffer< T >::get_size() const
{
	return size;
}
// -------------------------------------------------------------------------------------------- //
